/*
Copyright (c) 2014 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <ctype.h>
#include <stdlib.h>

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>

#include "x-basics.h"
#include "x-events.h"
#include "x-image.h"
#include "txtren.h"
#include "graymap.h"
#include "ptrbuf.h"
#include "resize_image.h"

#include "list.h"

#define DCLICK_DURATION  400

typedef struct { int x,y,w,h; } rect_t;

struct litem {
  uint32_t *image; uint32_t w,h;
  int y;
  void *ptr;
  struct litem *prev,*next;
  int selected;
};

struct xlist
{
  list_item_t head;
  int yoffset;
  Window window;
  ximage_t *image;  
  int w,h;
  int ch;
  rect_t upb, dnb;

  int pressed;
  int visible_buttons;

  void (*item_click)(void *, list_t *, list_item_t *);
  void (*item_dclick)(void *, list_t *, list_item_t *);
  void *arg;

  list_item_t *last_clicked_item;
  Time         last_click_time;
};


static uint32_t LI_A= 3;
static uint32_t LI_B= 3;
static uint32_t LI_C= 3;
static uint32_t LI_D= 16;
static uint32_t LI_E= 2;
static uint32_t LI_F= 3;
static uint32_t LI_fg= 0x0;
static uint32_t LI_bg= 0xededed;
static uint32_t LI_bgi= 0xffffff;
static uint32_t LI_G= 3;
static uint32_t LI_H= 3;
static uint32_t LI_J= 3;

static u_font_t *list_font;
static char * LI_font_name= ":default";
static int  LI_font_size= 12;

#include "fsel-uparr.c"
#include "fsel-dnarr.c"

list_item_t *list_item_new(uint32_t *idata, uint32_t iw, uint32_t ih,
                           const char *txt, void *ptr)
{
  uint8_t *tdata; uint32_t tw, th;
  list_item_t *li;
  uint32_t *image, sw, sh;
  uint32_t *rdata, rw, rh;

  if (!list_font)
    list_font= font_new(LI_font_name,LI_font_size);

  th= list_font->height;
  text_render(list_font, txt, &tw, &tdata);

  sw= tw + LI_D + LI_A + LI_B;
  sh= LI_E + list_font->height + LI_E;
  image= malloc(sizeof(uint32_t)*sw*sh);
  fill_rect(image, sw, sh, 0, 0, sw, sh, LI_bgi);

  graymap_draw(image, sw, sh, LI_A+LI_D+LI_B, LI_E,
               tdata, tw, th, 0, 0, LI_fg, LI_bgi,  tw, th); 
  free(tdata);

  if (!idata || !iw || !ih) goto no_icon;
  if (iw<=LI_D && ih<=LI_D) { rdata= idata; rw= iw; rh= ih; }
  else if (resize_image_argb(idata, iw, ih, LI_D, LI_D, &rdata, &rw, &rh))
         goto no_icon;

  blend_alpha_image(image, sw, sh, LI_A+(LI_D-rw)/2, (sh-rh)/2,
                    rdata, rw, rh, 0, 0,
                    rw, rh, LI_bgi);

  if (rdata!=idata) free(rdata);
no_icon:

  li= malloc(sizeof(*li));
  li->image= image; li->w= sw; li->h= sh; li->ptr= ptr;
  li->y= 0;
  li->prev= li->next= li;
  li->selected= 0;
  return li;
}

static void list_expose(list_t *L, XExposeEvent *ev);
static void list_press(list_t *L, XButtonEvent *ev);
static void list_release(list_t *L, XButtonEvent *ev);
static void list_destroy(list_t *L, XEvent *ev);

static void list_render(list_t *L);
static void scroll_down(list_t *L);
static void scroll_up(list_t *L);
static void hilight_rect(list_t *L, rect_t *R);


list_t *list_new(Window parent, int x,int y, int w,int h,
                 void *item_click_func, void *item_dclick_func, void *arg)
{
  list_t *L;
  XSetWindowAttributes attribs;

  L= calloc(1,sizeof(*L));
  L->head.next= L->head.prev= &L->head;

  attribs.background_pixmap= None;
  L->window= XCreateWindow(x_display, parent, x,y, w,h,
                            0, x_depth, InputOutput, x_visual,
                            CWBackPixmap, &attribs);

  L->image= create_image(w,h);

  L->w= w; L->h= h; 

  L->yoffset= 0;
  L->ch= L->h-2*LI_F;

  L->item_click= item_click_func;
  L->item_dclick= item_dclick_func;
  L->arg= arg;

  XSelectInput(x_display, L->window, ExposureMask | ButtonPressMask |
                                     StructureNotifyMask 
                                     | ButtonReleaseMask);
  register_event_handler(L->window, Expose, L, list_expose, NULL);
  register_event_handler(L->window, ButtonPress, L, list_press, NULL);
  register_event_handler(L->window, ButtonRelease, L, list_release, NULL);
  register_event_handler(L->window, DestroyNotify, L, list_destroy, NULL);

  list_render(L);

  L->upb.x= LI_G;
  L->upb.y= (L->h-bi_uparrsmall_height-bi_dnarrsmall_height-LI_J)/2;
  if (L->upb.y<0) L->upb.y=0;
  L->upb.w= bi_uparrsmall_width;
  L->upb.h= bi_uparrsmall_height;

  L->dnb.x= L->upb.x;
  L->dnb.y= L->upb.y + LI_J + L->upb.h;
  L->dnb.w= bi_dnarrsmall_width;
  L->dnb.h= bi_dnarrsmall_height;

  XMapWindow(x_display, L->window);
  return L;
}



static void list_expose(list_t *L, XExposeEvent *ev)
{
  put_image(L->window, ev->x, ev->y,
            L->image, ev->x, ev->y,
            ev->width, ev->height);
}

static int in_rect(XButtonEvent *ev, rect_t *R)
{
  int x= ev->x, y= ev->y;
  if (x<R->x) return 0;
  if (y<R->y) return 0;
  if (x>=R->x+R->w) return 0;
  if (y>=R->y+R->h) return 0;
  return 1;
}

static void list_press(list_t *L, XButtonEvent *ev)
{
  int ca_x;
  list_item_t *li;
  int y;

  if (ev->button==Button4) { scroll_up(L); return ; }
  if (ev->button==Button5) { scroll_down(L); return ; }
  if (ev->button!=Button1) return ;
  if ((L->visible_buttons &1) && in_rect(ev, &L->upb)) 
       { L->pressed= 1; hilight_rect(L,&L->upb); return ; } 
  if ((L->visible_buttons &2) && in_rect(ev, &L->dnb))
       { L->pressed= 2; hilight_rect(L,&L->dnb); return ; }

  L->pressed= 0;
  ca_x= LI_G+LI_H+bi_uparrsmall_width;
  if (ev->x<ca_x) return ;
  y= ev->y + L->yoffset + LI_G;
  for(li=L->head.next;li!=&L->head;li=li->next)
    if (y>=li->y && y<li->y+li->h) 
      break;

   if (li==&L->head) return ;
   if (li==L->last_clicked_item && 
        (ev->time-L->last_click_time<=DCLICK_DURATION))
   {
     L->last_click_time= 0;
     L->last_clicked_item= NULL;
     if (L->item_dclick) (*L->item_dclick)(L->arg, L, li);
     return ;
   }
   if (L->item_click) (*L->item_click)(L->arg, L, li);
   L->last_clicked_item= li;
   L->last_click_time= ev->time;
}

static void list_release(list_t *L, XButtonEvent *ev)
{
  if (ev->button!=Button1) return ;
  switch(L->pressed)
  {
  case 1: scroll_up(L); break;
  case 2: scroll_down(L); break;
  }
  L->pressed= 0;
}

static void hilight_rect(list_t *L, rect_t *R)
{
  xor_rect(L->image->data, L->w, L->h, R->x, R->y, R->w, R->h, 0xffffff);
  XClearArea(x_display, L->window, R->x, R->y, R->w, R->h, True);
}

static void list_item_render(list_t *L, list_item_t *li,int expose)
{
  int ca_x;
  int sW, eW;
  int sI, eI;
  int cw;

  ca_x= LI_G+LI_H+bi_uparrsmall_width;

  sW= L->yoffset; eW= L->yoffset+L->ch;

    sI= li->y; eI= li->y+li->h;
    if (sI>=eW || eI<=sW) return;

    if (sI<sW) sI= sW;
    if (eI>eW) eI= eW;

    if (sI>=eI) return;
    cw= L->w - LI_C - ca_x;
    if (cw>li->w) cw= li->w;

    copy_area(L->image->data, L->w, L->h, ca_x, sI-L->yoffset, 
              li->image, li->w, li->h,    0, sI-li->y, 
                cw, eI-sI); 

    if (li->selected)
      xor_rect(L->image->data, L->w, L->h, ca_x, sI-L->yoffset,
                                           cw, eI-sI, 0xffffff);

  if (expose)
    XClearArea(x_display, L->window, ca_x, sI-L->yoffset,
                                     cw, eI-sI, True);
}

static void list_render(list_t *L)
{
  list_item_t *li;
  int ca_x;

  ca_x= LI_G+LI_H+bi_uparrsmall_width;

  fill_rect(L->image->data, L->w, L->h, 0,0,   ca_x,L->h, LI_bg);
  fill_rect(L->image->data, L->w, L->h, ca_x, 0, L->w-ca_x, L->h, LI_bgi);

  L->visible_buttons= 0;
  L->pressed= 0;

  if (L->head.next==&L->head) return ;

  for(li=L->head.next;li!=&L->head;li=li->next)
     list_item_render(L, li,0);

  if (L->head.next->y<L->yoffset)
  {
    blend_alpha_image
      (L->image->data, L->w, L->h, L->upb.x, L->upb.y,
       (uint32_t*) bi_uparrsmall, bi_uparrsmall_width, bi_uparrsmall_height,
             0,0, 
              bi_uparrsmall_width, bi_uparrsmall_height, LI_bg);
    L->visible_buttons |= 1;
  }

  if (L->head.prev->y+L->head.prev->h>L->yoffset+L->ch)
  {
   blend_alpha_image(L->image->data, L->w, L->h, L->dnb.x, L->dnb.y,
         (uint32_t*) bi_dnarrsmall, bi_dnarrsmall_width, bi_dnarrsmall_height,
                    0,0, 
                     bi_dnarrsmall_width, bi_dnarrsmall_height, LI_bg);
    L->visible_buttons |= 2;
  }
}

static void update_list(list_t *L)
{
  list_render(L);
  XClearArea(x_display, L->window, 0,0, L->w, L->h, True);
}

void list_append(list_t *L, list_item_t *li)
{
  li->prev= L->head.prev;
  li->next= &L->head;
  li->selected= 0;
  L->head.prev->next= li;
  L->head.prev= li;
  if (li->prev==&L->head) li->y= 0;
                    else  li->y= li->prev->y + li->prev->h;
  update_list(L);
}

static void scroll_up(list_t *L)
{
  list_item_t *first,*li;
  first= L->head.next;
  if (first==&L->head) return ;
  if (first->y>=L->yoffset) return ;

  for(li=first;li!=&L->head;li=li->next)
    if (li->y<=L->yoffset && li->y+li->h>L->yoffset) 
       break;
  L->yoffset= li->y+li->h;
  L->yoffset-= L->ch;
  if (L->yoffset<0) L->yoffset= 0;
  list_render(L);
  XClearArea(x_display, L->window, 0,0, L->w, L->h, True);
}

static void scroll_down(list_t *L)
{
  list_item_t *last,*li;
  last= L->head.prev;
  if (last==&L->head) return ;
  if (last->y+last->h<=L->yoffset+L->ch) return ;

  for(li=L->head.next;li!=&L->head;li=li->next)
    if (li->y<L->yoffset+L->ch && li->y+li->h>=L->yoffset+L->ch) 
       break;

  if (li->y+li->h==L->yoffset+L->ch) 
  {
     if (li==last) return ;
     li= li->next;
  }
  L->yoffset= li->y;
  list_render(L);
  XClearArea(x_display, L->window, 0,0, L->w, L->h, True);
}

void list_item_select(list_t *L, list_item_t *li,int sel)
{
  if (li->selected==sel) return ;
  li->selected= sel;
  list_item_render(L,li,1);
}

static void update_y_coords(list_t *L)
{
  list_item_t *li;
  int y;
  y= 0;
  for(li=L->head.next;li!=&L->head;li=li->next)
  {
    li->y= y;
    y+= li->h;
  }
}

void list_prepend(list_t *L, list_item_t *li)
{
  li->next= L->head.next;
  li->prev= &L->head;

  L->head.next->prev= li;
  L->head.next= li;
  li->selected= 0;

  update_y_coords(L);
  update_list(L);
}

void list_insert_before(list_t *L, list_item_t *anchor, list_item_t *li)
{
  li->prev= anchor->prev;
  li->next= anchor;

  anchor->prev->next= li;
  anchor->prev= li;
  li->selected= 0;
  
  update_y_coords(L);
  update_list(L);
}

void list_insert_after(list_t *L, list_item_t *anchor, list_item_t *li)
{
  li->next= anchor->next;
  li->prev= anchor;

  anchor->next->prev= li;
  anchor->next= li;
  li->selected= 0;
  
  update_y_coords(L);
  update_list(L);
}

void list_remove(list_t *L, list_item_t *li)
{
  int end;
  list_item_t *last,*k;
  li->prev->next= li->next;
  li->next->prev= li->prev;

  update_y_coords(L);
  last= L->head.prev;
  if (last==&L->head) goto from_first; 
  if (L->yoffset<=last->y+last->h) goto done;

  end= last->y+last->h;
  for(k=last->prev;k!=&L->head;k=k->prev)
     if (end-k->y >= L->ch) break;
  if (k==&L->head) goto from_first;

  L->yoffset= end - L->ch;
  if (L->yoffset<0) goto from_first;

done:
  update_list(L);
  return ;
from_first:
  L->yoffset= 0;
  goto done;
}

void *list_item_data(list_item_t *li)
{
  return li->ptr;
}

static void list_delete_all(list_t *L)
{
  list_item_t *li, *N;
  for(li=L->head.next;li!=&L->head;li=N)
  {
    N= li->next;
    free(li->image);
    free(li);
  }
  L->head.next= L->head.prev= &L->head;
}

void list_clear(list_t *L)
{
  list_delete_all(L);
  L->yoffset= 0;
  list_render(L);
  XClearArea(x_display, L->window, 0,0, L->w, L->h, True);
}

void list_item_free(list_item_t *li)
{
  if (li->image) free(li->image);
  free(li);
}

list_item_t *list_find_item(list_t *L, void *ptr)
{
  list_item_t *li; 
  for(li=L->head.next;li!=&L->head;li=li->next)
     if (li->ptr==ptr) return li;
  return NULL;
}

static void list_destroy(list_t *L, XEvent *ev)
{
  list_delete_all(L);
  destroy_image(L->image);
  unregister_event_handler(L->window, Expose, NULL);
  unregister_event_handler(L->window, ButtonPress, NULL);
  unregister_event_handler(L->window, ButtonRelease, NULL);
  unregister_event_handler(L->window, DestroyNotify, NULL);
  free(L);
}

void list_free(list_t *L)
{
  XDestroyWindow(x_display, L->window);
  list_destroy(L, NULL);
}

