/*
Copyright (c) 2014 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/* This file implements a resizing algorithm for reducing the size of
   an image to fit in a given box. 
*/
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "resize_image.h"

static double rgb2linear(uint32_t K)
{
  return pow((double) K  / 255.0,2.2);
}

static uint32_t linear2rgb(double V)
{
  if (V<=0.0) return 0;
  if (V>=1.0) return 255;
  return pow(V,1/2.2)*255;
}

static void resize_channel
     (uint32_t *src, uint32_t width, uint32_t height,
      uint32_t *dst, uint32_t owidth, uint32_t oheight,
      uint32_t shift, double *row, double *lrow, double weight)
{
  int r,c;
  int X,Y;
  int OR,OP; double row_weight, col_weight;

  double row_total;
  double pixel_total, pixel;

  row_total= 0.0;
  for(c=0;c<width;c++) row[c]= 0;

  Y=0;

  for(r=0;r<height;r++)
  {
    for(c=0;c<width;c++)
      lrow[c] = rgb2linear((src[r*width+c] >> shift) & 0xff);

    if (row_total+weight>=1.0) { OR=1; row_weight= 1.0- row_total; }
                          else { OR= 0; row_weight= weight; }

    for(c=0;c<width;c++) row[c] += lrow[c] * row_weight;

    row_total+= row_weight;
    if (!OR) continue;
    
    X= 0;
    pixel_total= 0.0;
    pixel= 0.0;
    for(c=0;c<width;c++)
    {
       if (pixel_total+weight>=1.0) { OP= 1; col_weight= 1.0 - pixel_total; }
                               else { OP= 0; col_weight= weight; }
       pixel += row[c]* col_weight;
       pixel_total += col_weight;
       if (!OP) continue;

       dst[Y*owidth+X] |= linear2rgb(pixel) << shift;
       X++;
       pixel_total = weight - col_weight;
       pixel= row[c]* pixel_total;
    }
    Y++; 
    row_total= weight - row_weight;
    for(c=0;c<width;c++) row[c] = lrow[c] * row_total;
  }
}

int resize_image_argb
    (uint32_t *data, uint32_t width, uint32_t height, 
     uint32_t target_width, uint32_t target_height,
     uint32_t **Rdata, uint32_t *Rwidth, uint32_t *Rheight)
{
  double aspect;
  double *row;
  uint32_t owidth, oheight;
  uint32_t *odata;

  double weight, total;
  int c;

  if (width==0 || height==0 || target_width==0 || target_height==0)
    return 1;
  if (target_width>=width && target_height>=height) return 1;

  aspect= (double) height / width;
  
  if (aspect*target_width<=target_height)
  {
     owidth=  target_width;
     oheight= target_width * aspect;
  } 
  else
  {
     oheight= target_height;
     owidth= target_height / aspect;
  }

  weight= (double) owidth / width;

  /* The above calculations aren't very good. The actual resizing 
     algorithm will do the following calculations and finally reach
     a smaller width or height (it depends on how you calculate the
     weight). This is caused by rounding errors, I guess. The weight
     can't be exactly represented in floating point so the actual 
     weight we use is a little smaller than necessary. This causes
     us to reach a slightly lower height when we add so many weights
     together... */
  total= 0.0; owidth= 0;
  for(c=0;c<width;c++)
     if (total+weight>=1.0) { owidth++; total= weight - (1-total); } 
                       else { total+= weight; }

  total= 0.0; oheight= 0;
  for(c=0;c<height;c++)
     if (total+weight>=1.0) { oheight++; total= weight - (1-total); } 
                       else { total+= weight; }

  if (owidth<1 || oheight<1) return 1;

  odata= malloc(sizeof(uint32_t)*owidth*oheight);

  /* The second half of the row array will be used for temporarily
     holding the linear values corresponding to the current row. 
     The first half is the accumulator.. */
  row= malloc(2*sizeof(double)*width);

  *Rwidth= owidth; *Rheight= oheight; *Rdata= odata;

  memset(odata, 0, sizeof(uint32_t)*owidth*oheight);

  resize_channel(data, width, height, odata, owidth, oheight,
                 0, row, row+width, weight);
  resize_channel(data, width, height, odata, owidth, oheight,
                 8, row, row+width, weight);
  resize_channel(data, width, height, odata, owidth, oheight,
                 16, row, row+width, weight);
  resize_channel(data, width, height, odata, owidth, oheight,
                 24, row, row+width, weight);
  
  free(row);
  return 0;
}


