/*
Copyright (c) 2014 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>

#include "x-basics.h"
#include "x-props.h"
#include "x-events.h"
#include "x-misc-props.h"
#include "x-image.h"
#include "txtren.h"
#include "config.h"
#include "clients.h"
#include "taskbar.h"
#include "menu.h"
#include "graymap.h"

/****************************************************************
  Each taskbar item is repsented by the following struct. We
  use this for activating windows or displaying the window menu.
****************************************************************/
typedef struct
{
  client_window_t *client;
  int x;
  int w;
} tb_item_t;

/****************************************************************
  I will display only a handful of items at a time. Having
  more than 100 windows on a desktop doesn't make sense.
****************************************************************/
#define MAX_ITEMS 100
static tb_item_t items[MAX_ITEMS];
static int       n_items;
static int       pressed_item=-1;
static int       pressed_button=-1;

/****************************************************************
  We display the following menu when an item is right clicked.
  It gets re-created for different desktops. The second variable
  holds for which desktop the current menu (if any) was created
  for. 
****************************************************************/
static menu_t   *item_menu;
static int       item_menu_desk=-1;

static ximage_t *tb_image;
static Window    tb_window;
static int       tb_w, tb_h;
static u_font_t *tb_font;
static int       tb_parent_y;
static int       tb_parent_x;

static uint32_t  TB_bg        = 0xededed;
static uint32_t  TB_fg        = 0x0;
static uint32_t  TB_icon_size = 16;
static char     *TB_font_name = ":default";
static int       TB_font_size = 10;
static int       TB_max_size  = 150;
static int       TB_border    = 0x909090;
static int       TB_active    = 0x0;
static int       TB_left      = 3;       // left padding
static int       TB_mid       = 2;       // space between icon and text
static int       TB_right     = 3;       // right padding

static uint32_t  c_desk;
static uint32_t  n_desk;
static Window    active_window;


static void tb_release(void *ignore, XButtonEvent *ev);
static void tb_click(void *ignore, XButtonEvent *ev);
static void tb_expose(void *ignore, XExposeEvent *ev);


static void render_taskbar();
static void render_taskbar_item(client_window_t *cw, int x,int w);
static void draw_active_indicator(int erase);

static void item_menu_click(tb_item_t *it, XEvent *ev);
static void create_item_menu();
static void display_item_menu(tb_item_t *it);

void init_taskbar
  (Window parent, int parent_x, int parent_y, int width, int height,
                  cfg_taskbar_t *cfg,Window *Rwindow)
{
  tb_image= create_image(width, height);
  tb_w= width; tb_h= height;
  tb_window= create_window_Ob(parent, 0,0, tb_w, tb_h,
                                 ButtonReleaseMask |
                                 ButtonPressMask | ExposureMask, 0);
  tb_parent_y= parent_y; 
  *Rwindow= tb_window;

  if (cfg->G_bg) TB_bg= cfg->bg;
  if (cfg->G_fg) TB_fg= cfg->fg;
  if (cfg->G_border) TB_border= cfg->border;
  if (cfg->G_active) TB_active= cfg->active;
  if (cfg->G_font_size) TB_font_size= cfg->font_size;
  if (cfg->G_font_name) TB_font_name= cfg->font_name;
  if (cfg->G_max_size)  TB_max_size= cfg->max_size;
  if (cfg->G_icon_size)  TB_icon_size= cfg->icon_size;
  if (cfg->G_right)      TB_right= cfg->right;
  if (cfg->G_left)      TB_left= cfg->left;
  if (cfg->G_mid)      TB_mid= cfg->mid;
 
  tb_font= font_new(TB_font_name, TB_font_size);

  register_event_handler(tb_window, Expose, NULL, tb_expose, NULL);
  register_event_handler(tb_window, ButtonPress, NULL, tb_click, NULL);
  register_event_handler(tb_window, ButtonRelease, NULL, tb_release, NULL);

  render_taskbar();
}

static void render_taskbar()
{
  client_window_t *cw;
  int i, x,isize;
  Window *windows;
  size_t n_windows;

  fill_rect(tb_image->data, tb_w, tb_h, 0,0, tb_w, tb_h, TB_bg);


  if (get_window_property_windows(x_root_window, a_net_client_list,
                                   &n_windows, &windows, 0))
     return ;

  if (!n_windows) goto done;

  if (get_current_desktop(&c_desk)) c_desk=0;

  n_items= 0;
  for(i=0;i<n_windows;i++)
  {
    if (n_items>=MAX_ITEMS) break;
    cw= get_client(windows[i]);
    if (get_client_type(cw) || cw->type!=0) continue;
    if (!get_client_desktop_number(cw) && cw->desk!=c_desk) continue;
    items[n_items].client= cw;
    n_items++;
  }

  if (n_items==0) goto done;
  isize= tb_w / n_items - 1;
  while(n_items>4 && isize<5)
  {
     n_items--; 
     isize= tb_w / n_items - 1;
  }

  if (isize>TB_max_size) isize= TB_max_size;

  x= 0;
  for(i=0;i<n_items && x<tb_w;i++)
  {
    render_taskbar_item(items[i].client, x,isize);
    items[i].x= x;
    items[i].w= isize;
    x+= isize + 1;
  }

  draw_active_indicator(0);
done:
  free(windows);
}

static void render_taskbar_item(client_window_t *cw, int x,int w)
{
  int vpad, ix,iy, iw, ih,rem,ty;
  char *str;
  uint32_t titlew;
  uint8_t *title;

  vpad= (tb_h - TB_icon_size)/2;

  if (get_client_icon(cw, TB_icon_size, TB_icon_size))
     goto noicon;

  ix= iy= 0;
  iw= cw->icon_width;
  ih= cw->icon_height;
  if (iw>TB_icon_size) 
     {iw= TB_icon_size; ix= (cw->icon_width-TB_icon_size/2);}
  if (ih>TB_icon_size) 
     {ih= TB_icon_size; iy= (cw->icon_height-TB_icon_size/2);}


  blend_alpha_image(tb_image->data, tb_w, tb_h, x + TB_left+1, vpad+1,
                    cw->icon_data, cw->icon_width, cw->icon_height, ix, iy,
                    iw, ih, TB_bg);
noicon:
  if (get_client_title(cw)) str= "Window"; else str= cw->title;
  text_render(tb_font, str, &titlew, &title);
  rem= w - TB_left - TB_right - TB_mid - TB_icon_size - 2;
  ty= (tb_h - tb_font->height)/2;
  if (ty<0) ty= 0;
  graymap_draw(tb_image->data, tb_w, tb_h,
               x+ TB_left + TB_icon_size+ TB_mid + 1, ty,
               title, titlew, tb_font->height, 0,0,
                TB_fg, TB_bg, rem, tb_font->height);
  free(title);

  draw_rect(tb_image->data, tb_w, tb_h, x, 0, w, tb_h, TB_border);
}

static void hilight_item(int ino)
{
  if (ino<0 || ino>=n_items) return ;
  xor_rect(tb_image->data, tb_w, tb_h,
           items[ino].x, 0, items[ino].w, tb_h, 0xffffff);  
  XClearArea(x_display,tb_window, items[ino].x, 0, items[ino].w, tb_h, True);
}

static void tb_release(void *ignore, XButtonEvent *ev)
{
  if (pressed_item<0 || pressed_item>=n_items) return ;
  hilight_item(pressed_item);
  if (pressed_button==Button1)
  {
    XEvent event;

    XRaiseWindow(x_display, items[pressed_item].client->id);

    event.xclient.type = ClientMessage;
    event.xclient.window = items[pressed_item].client->id;
    event.xclient.message_type= a_net_active_window;
    event.xclient.format = 32;
    event.xclient.data.l[0] = 2;
    event.xclient.data.l[1] = 0;
    event.xclient.data.l[2] = 0;
    event.xclient.data.l[3] = 0;
    event.xclient.data.l[4] = 0;

    XSendEvent(x_display, x_root_window, False,
                 SubstructureNotifyMask | SubstructureRedirectMask, &event);
  }
  else
  {
    display_item_menu(items+pressed_item);
  }
  pressed_item= -1;
}

static void tb_click(void *ignore, XButtonEvent *ev)
{
  int i;

  if (ev->button!=Button1 && ev->button!=Button3) return ;

  for(i=0;i<n_items;i++)
    if (ev->x>=items[i].x && ev->x<items[i].x+items[i].w) break;

  if (i==n_items) { pressed_item= -1; return ; }
  pressed_button= ev->button;
  pressed_item= i;
  hilight_item(i);
}

static void tb_expose(void *ignore, XExposeEvent *ev)
{
  put_image(tb_window, ev->x, ev->y,
            tb_image, ev->x, ev->y,
            ev->width, ev->height);
}

void taskbar_current_desktop_changed()
{
  render_taskbar();
  XClearArea(x_display, tb_window, 0,0, tb_w, tb_h,True);
}

void taskbar_client_list_changed()
{
  render_taskbar();
  XClearArea(x_display, tb_window, 0,0, tb_w, tb_h,True);
}

void taskbar_number_of_desktops_changed()
{
  item_menu_desk= -1;
}

static tb_item_t *find_taskbar_item(Window window)
{
  int i;
  for(i=0;i<n_items;i++)
     if (items[i].client->id==window) return items+i;
  return NULL;
}

static void draw_active_indicator(int erase)
{
  tb_item_t *it;
  size_t n;
  Window *windows;

  if (!erase || active_window==None) goto fresh;

  it= find_taskbar_item(active_window);
  if (!it) goto fresh;

  draw_rect(tb_image->data, tb_w, tb_h, it->x+1, 1, it->w-2, tb_h-2, TB_bg);
  draw_rect(tb_image->data, tb_w, tb_h, it->x+2, 2, it->w-4, tb_h-4, TB_bg);
  XClearArea(x_display, tb_window, it->x+1, 1, it->w-2, tb_h-2, True);
fresh:
  active_window= None;
  if (get_window_property_windows(x_root_window, a_net_active_window,
                                   &n, &windows, 1) || n==0)
     return ;
  active_window= windows[0];
  free(windows); 
  it= find_taskbar_item(active_window);
  if (!it) return ;

  draw_rect(tb_image->data, tb_w, tb_h, it->x+1, 1, it->w-2, tb_h-2,
                                                                TB_active);
  draw_rect(tb_image->data, tb_w, tb_h, it->x+2, 2, it->w-4, tb_h-4,
                                                                TB_active);
  XClearArea(x_display,tb_window, it->x+1, 1, it->w-2, tb_h-2, True);
}

void taskbar_active_window_changed()
{
  draw_active_indicator(1);
}

void taskbar_resize(int width, int height)
{
  destroy_image(tb_image);
  tb_image= create_image(width, height);
  tb_w= width; tb_h= height;
  XResizeWindow(x_display,tb_window, width, height);
  render_taskbar();
}

static void display_item_menu(tb_item_t *it)
{
  Window Mw; unsigned int Me;
  if (item_menu) 
  {
    if (!menu_is_released(item_menu)) return ;
    if (item_menu_desk!=c_desk)
    {
       menu_free(item_menu);
       item_menu= NULL;
       item_menu_desk= -1;
    }
  }
  int self_x, self_y;
  unsigned int dummy_u; Window dummy_w;

  if (!item_menu) create_item_menu();
  XGetGeometry(x_display, tb_window, &dummy_w, &self_x, &self_y,
                          &dummy_u, &dummy_u, &dummy_u, &dummy_u);
  menu_display_top(item_menu,x_root_height-tb_parent_y,
                    self_x+it->x + tb_parent_x, &Mw, &Me);
  register_event_handler(Mw, Me, it, item_menu_click, NULL);
}

static void create_item_menu()
{
  int i;
  item_menu= menu_start();
  if (get_number_of_desktops(&n_desk)) n_desk=1;
  for(i=0;i<n_desk;i++)
    if (i!=c_desk)
    {
       char buf[200];
       sprintf(buf,"Move to [%d]",i+1);
       menu_add_item(item_menu, 0,0, NULL, buf, 1000 + i);
    }
  menu_add_item(item_menu, 0,0, NULL, "Close", 1);
  menu_end(item_menu);
  item_menu_desk= c_desk;
}

static void item_menu_click(tb_item_t *it, XEvent *ev)
{
  int k;
  unregister_event_handler(ev->xany.type, ev->xany.window,NULL);
  if (!item_menu) return ;
  menu_release(item_menu);
  if (it>=items+n_items) return ;
  k= menu_get_selected(item_menu);
  if (k==1)
  {
    request_close_window(it->client->id);
    // printf("closing window %d\n", it->client->id);
  }
  else  if (k>=1000 && k<2000)
  {
    k-= 1000;
    change_window_desktop(it->client->id, k);
    //printf("sending window %d to workspace %d\n", it->client->id, k);
  }
}

void taskbar_client_desktop_changed(Window window)
{
  render_taskbar();
  XClearArea(x_display, tb_window, 0,0, tb_w, tb_h,True);
}

void taskbar_window_title_changed(Window window)
{
  render_taskbar();
  XClearArea(x_display, tb_window, 0,0, tb_w, tb_h,True);
}
